declare namespace CloudbaseIndexChecker {
  export const getIndexSuggestion: (options: IGetIndexSuggestionOptions) => IIndexSuggestion
  export const getQueryKeyPattern: (spec: IQuerySpec) => Record<string, any>
  export const transpile: (options: ITranspileAction) => string

  interface IGetIndexSuggestionOptions {
    spec: IQuerySpec
    indices: IIndexInfo[]
  }
  
  interface IIndexSuggestion {
    // normal: any[]
    perfect: IPerfectCandidate[]
    warnings: IIndexWarning[]
  }
  
  interface IPerfectCandidate {
    keys: IKeyIndexInfo[]
    samePrefixIndices: IIndexInfo[]
  }
  
  interface IIndexWarningFullScan {
    type: 'FULL_SCAN'
    reason: 'EMPTY_QUERY' | 'AGGREGATE_FIRST_STAGE_NOT_MATCH'
  }
  
  interface IIndexWarningRegExp {
    type: 'REGEXP'
    field: string
    reason: 'CASE_INSENSITIVE' | 'NOT_PREFIX_EXPR'
  }
  
  interface IIndexWarningUnselectiveOperator {
    type: 'UNSELECTIVE_OPERATOR'
    field: string
    operator: string
  }
  
  type IIndexWarning = IIndexWarningFullScan | IIndexWarningRegExp | IIndexWarningUnselectiveOperator
  
  interface IQuerySpec {
    query: Record<string, any>
    sort?: {
      field: string
      type: 1 | -1
    }[]
    projection?: Record<string, number>
  }
  
  interface IOrder {
    field: string
    direction: ORDER_DIRECTION
  }
  
  enum ORDER_DIRECTION {
    ASC = 0,
    DESC = 1,
  }
  
  interface IIndexInfo {
    name: string
    keys: IKeyIndexInfo[]
  }
  
  interface IKeyIndexInfo {
    field: string
    type: 1 | -1 | '2dsphere'
  }

  // transpile
  type ITranspileAction =
    ITranspileActionGet |
    ITranspileActionSet |
    ITranspileActionUpdate |
    ITranspileActionAdd |
    ITranspileActionRemove |
    ITranspileActionCount |
    ITranspileActionAggregate |
    ITranspileActionQueryClauseOnly |
    ITranspileActionUpdateDataOnly

  interface ITranspileActionBase {
    style?: 'default' | 'condensed'
    collection: string
    action: DBScriptAction
    query?: Record<string, any>
    data?: Record<string, any>
  }

  interface ITranspileActionGet extends ITranspileActionBase {
    action: 'get'
    query: Record<string, any>
    skip?: number
    limit?: number
    projection?: Record<string, 1 | 0>
    order?: IOrderDesc[]
    isTransaction?: boolean
  }

  interface ITranspileActionSet extends ITranspileActionBase {
    action: 'set'
    _id: string | number
    data: Record<string, any>
    isTransaction?: boolean
  }

  interface ITranspileActionUpdate extends ITranspileActionBase {
    action: 'update'
    data: Record<string, any>
    multi?: boolean
    isTransaction?: boolean
  }

  interface ITranspileActionAdd extends ITranspileActionBase {
    action: 'add'
    query?: never
    list: Record<string, any>[]
    isTransaction?: boolean
  }

  interface ITranspileActionRemove extends ITranspileActionBase {
    action: 'remove'
    query: Record<string, any>
    data?: never
    multi?: boolean
    isTransaction?: boolean
  }

  interface ITranspileActionCount extends ITranspileActionBase {
    action: 'count'
    data?: never
  }

  interface ITranspileActionAggregate extends ITranspileActionBase {
    action: 'aggregate'
    collection: string
    stages: Record<string, any>[]
  }

  interface ITranspileActionQueryClauseOnly extends ITranspileActionBase {
    action: '__query__'
    query: Record<string, any>
  }

  interface ITranspileActionUpdateDataOnly extends ITranspileActionBase {
    action: '__update_data__'
    data: Record<string, any>
  }

  type DBScriptAction = 'get' | 'set' | 'update' | 'add' | 'remove' | 'count' | 'aggregate' |
    '__query__' | '__update_data__'

  interface IOrderDesc {
    field: string
    direction: 1 | -1
  }

  type IBuilderStyle = 'default' | 'condensed'

  interface IBuilderContext {
    tabSize: number
    newLine: string
    gap: string
  }
}

export = CloudbaseIndexChecker
